#!/usr/bin/python3 -i
#
# Copyright (c) 2019-2020 Valve Corporation
# Copyright (c) 2019-2024 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from khronos_base_generator import write


class KhronosStructHandleMappersBodyGenerator():
    """Base class for generating struct handle mappers body code."""

    # Recursively search a structs members to see if they too belong in the
    # output struct list.  This could be because an including struct is an
    # output struct.
    def process_struct_members_to_output_struct(self, value):
        for member in self.all_struct_members[value.base_type]:
            if (
                self.is_struct(member.base_type)
                and not self.is_struct_black_listed(member.base_type)
                and (member.base_type in self.structs_with_handles)
                and (member.base_type not in self.output_structs)
            ):
                self.output_structs.append(member.base_type)
                self.process_struct_members_to_output_struct(member)

    def write_struct_handle_wrapper_content(self):
        for struct in self.get_all_filtered_struct_names():
            if struct in self.STRUCT_MAPPERS_BLACKLIST:
                continue

            child_has_handles = self.child_struct_has_handles(struct)
            if ( child_has_handles or
                (struct in self.structs_with_handles) or
                (struct in self.GENERIC_HANDLE_STRUCTS)
            ):
                handle_members = list()
                generic_handle_members = dict()

                if struct in self.structs_with_handles:
                    handle_members = self.structs_with_handles[struct].copy()

                if struct in self.GENERIC_HANDLE_STRUCTS:
                    generic_handle_members = self.GENERIC_HANDLE_STRUCTS[struct
                                                                         ]

                # Determine if the struct only contains members that are structs that contain handles or static arrays of handles,
                # and does not need a temporary variable referencing the struct value.
                needs_value_ptr = False
                if generic_handle_members or child_has_handles:
                    needs_value_ptr = True
                else:
                    for member in handle_members:
                        if (
                            (self.is_handle_like(member.base_type)) and
                            not (member.is_array and not member.is_dynamic)
                        ) or (member.base_type in self.MAP_STRUCT_TYPE):
                            needs_value_ptr = True
                            break

                body = '\n'
                body += 'void MapStructHandles(Decoded_{}* wrapper, const CommonObjectInfoTable& object_info_table)\n'.format(
                    struct
                )
                body += '{\n'

                if not needs_value_ptr:
                    body += '    if (wrapper != nullptr)\n'
                    body += '    {'
                else:
                    body += '    if ((wrapper != nullptr) && (wrapper->decoded_value != nullptr))\n'
                    body += '    {\n'
                    body += '        {}* value = wrapper->decoded_value;\n'.format(
                        struct
                    )

                    # Add handling for parent/child structs since this actually might be one of the children.
                    if child_has_handles:
                        type_var = self.get_struct_type_var_name()
                        body += '\n'
                        body += f'        switch (value->{type_var})\n'
                        body += '        {\n'
                        body += '            default:\n'
                        body += '                // Handle as base-type below\n'
                        body += '                break;\n'

                        for child in self.children_structs[struct]:
                            if child not in self.structs_with_handles:
                                continue
                            switch_type = self.struct_type_names[child]

                            body += f'            case {switch_type}:\n'
                            body += f'                MapStructHandles(reinterpret_cast<Decoded_{child}*>(wrapper),\n'
                            body += f'                                 object_info_table);\n'
                            body += '                // Return here because we processed the appropriate data in\n'
                            body += '                // the correct structure type\n'
                            body += '                return;\n'
                        body += '        }\n'

                body += self.make_struct_handle_mappings(
                    struct, handle_members, generic_handle_members
                )
                body += '    }\n'
                body += '}'

                write(body, file=self.outFile)

        # Print out a function to handle mapping the extended struct types
        self.newline()
        next_field = self.get_extended_struct_var_name()
        next_var = next_field.lower()
        type_var = self.get_struct_type_var_name()
        cast_type = 'const {}MetaStructHeader*'.format(self.get_api_prefix())

        lines = []
        lines.append('void Map{func}StructHandles({node}Node* {var}, const CommonObjectInfoTable& object_info_table)'.format(
                func=self.get_extended_struct_func_prefix(),
                node=self.get_extended_struct_node_prefix(),
                var=next_var
            )
        )
        lines.append('{')
        lines.append(f'    while ({next_var})')
        lines.append('    {')
        lines.append(f'        void *wrapper = {next_var}->GetMetaStructPointer();')
        lines.append(f'        const auto* header = reinterpret_cast<{cast_type}>({next_var}->GetMetaStructPointer());')
        lines.append('')
        lines.append(f'        switch (*header->{type_var})')
        lines.append('        {')
        lines.append('        default:')
        lines.append(f'            // TODO: Report or raise fatal error for unrecognized {type_var}?')
        lines.append('            break;')
        write('\n'.join(lines), file=self.outFile)

        extended_list = []
        for struct in self.all_extended_structs:
            for ext_struct in self.all_extended_structs[struct]:
                if ext_struct not in extended_list and ext_struct not in self.all_struct_aliases:
                    extended_list.append(ext_struct)

        for base_type in sorted(extended_list):
            if base_type in self.structs_with_handles:
                write(
                    '        case {}:'.format(self.struct_type_names[base_type]),
                    file=self.outFile
                )
                write(
                    '            MapStructHandles(reinterpret_cast<Decoded_{}*>(wrapper), object_info_table);'
                    .format(base_type),
                    file=self.outFile
                )
                write('            break;', file=self.outFile)

        write('        }', file=self.outFile)
        write(f'        {next_var} = header->{next_field};', file=self.outFile)
        write('    }', file=self.outFile)
        write('}', file=self.outFile)

        # List of structs containing handles that are also used as output parameters for a command
        self.output_structs = []

        # Look for output structs that contain handles and add to list
        for cmd in self.get_all_filtered_cmd_names():
            for value_info in self.all_cmd_params[cmd][2]:
                if self.is_output_parameter(value_info) and self.is_struct(
                    value_info.base_type
                ) and (value_info.base_type in self.structs_with_handles
                       ) and (value_info.base_type not in self.output_structs):
                    self.output_structs.append(value_info.base_type)
                    self.process_struct_members_to_output_struct(value_info)

        # Generate handle adding functions for output structs with handles
        for struct in self.output_structs:
            self.newline()
            write(
                self.make_struct_handle_additions(
                    struct, self.structs_with_handles[struct]
                ),
                file=self.outFile
            )

        # Generate handle memory allocation functions for output structs with handles
        for struct in self.output_structs:
            if struct in self.structs_with_handle_ptrs:
                self.newline()
                write(
                    self.make_struct_handle_allocations(
                        struct, self.structs_with_handles[struct]
                    ),
                    file=self.outFile
                )

    def make_struct_handle_mappings(
        self, name, handle_members, generic_handle_members
    ):
        """Generating expressions for mapping struct handles read from the capture file to handles created at replay."""
        prefix_from_type = self.get_api_prefix_from_type(name)

        body = ''
        for member in handle_members:
            body += '\n'
            map_func = self.MAP_STRUCT_TYPE.get(member.base_type)

            if self.is_extended_struct_definition(member):
                func_id = self.get_extended_struct_func_prefix()
                body += f'        if (wrapper->{member.name})\n'
                body += '        {\n'
                body += f'            Map{func_id}StructHandles(wrapper->{member.name}, object_info_table);\n'
                body += '        }\n'
            elif map_func:
                if member.is_array:
                    body += '    if ({0})\n'\
                            '    {{\n'\
                            '        object_mapping::{2}(value->{0}->GetPointer(), {1}, {3});\n'\
                            '    }}\n'.format(
                        member.name, member.array_length, map_func[1], map_func[2]
                    )
                else:
                    if member.is_pointer:
                        body += '    if ({0})\n'\
                                '    {{\n'\
                                '        object_mapping::{1}(value->{0}->GetPointer(), {2});\n'\
                                '    }}\n'.format(
                            member.name, map_func[0], map_func[2]
                        )
                    else:
                        body += '        object_mapping::{}(value->{}, {});\n'.format(
                            map_func[0], member.name, map_func[2]
                        )
            elif self.is_struct(member.base_type):
                # This is a struct that includes handles.
                if member.is_array:
                    length_exprs = member.array_length.split(',')
                    length_count = len(length_exprs)

                    if member.pointer_count > 1 and length_count < member.pointer_count:
                        unwrap_function = f'MapStructPtrArrayHandles<Decoded_{member.base_type}*>'
                    else:
                        unwrap_function = f'MapStructArrayHandles<Decoded_{member.base_type}>'

                    body += '        {func}(wrapper->{name}->GetMetaStructPointer(), wrapper->{name}->GetLength(), object_info_table);\n'.format(
                        func=unwrap_function,
                        name=member.name
                    )
                elif member.is_pointer:
                    body += '        MapStructArrayHandles<Decoded_{}>(wrapper->{}->GetMetaStructPointer(), 1, object_info_table);\n'.format(
                        member.base_type, member.name
                    )
                else:
                    body += '        MapStructHandles(wrapper->{}, object_info_table);\n'.format(
                        member.name
                    )
            else:
                type = member.base_type
                prefix_from_type = self.get_api_prefix_from_type(member.base_type)
                func_id = member.base_type + 'Info'
                type = prefix_from_type + member.base_type[2:] + 'Info'
                object_info_table_get = ', &CommonObjectInfoTable::Get{}'.format(
                    func_id
                )

                # If it is an array or pointer, map with the utility function.
                if (member.is_array or member.is_pointer):
                    if member.is_dynamic or member.is_pointer:
                        body += '        value->{name} = handle_mapping::MapHandleArray<{type}>(&wrapper->{name}, object_info_table{});\n'.format(
                            object_info_table_get,
                            type=type,
                            name=member.name
                        )
                    else:
                        body += '        handle_mapping::MapHandleArray<{type}>(&wrapper->{name}, object_info_table{});\n'.format(
                            object_info_table_get,
                            type=type,
                            name=member.name
                        )
                else:
                    body += '        value->{name} = handle_mapping::MapHandle<{type}>(wrapper->{name}, object_info_table{});\n'.format(
                        object_info_table_get,
                        type=type,
                        name=member.name
                    )

        for member in generic_handle_members:
            body += '\n'
            body += '        value->{name} = handle_mapping::MapHandle(wrapper->{name}, value->{}, object_info_table);\n'.format(
                generic_handle_members[member],
                name=member
            )

        return body

    def make_struct_handle_additions(self, name, members):
        """Generating expressions for adding mappings for handles created at replay that are embedded in structs."""
        platform_type = self.get_api_prefix()

        body = 'void AddStructHandles(format::HandleId parent_id, const Decoded_{name}* id_wrapper, const {name}* handle_struct, CommonObjectInfoTable* object_info_table)\n'.format(
            name=name
        )
        body += '{\n'
        body += '    if (id_wrapper != nullptr)\n'
        body += '    {\n'

        for member in members:
            if self.is_extended_struct_definition(member):
                func_id = self.get_extended_struct_func_prefix()
                body += f'        if (id_wrapper->{member.name})\n'
                body += '        {\n'
                body += f'            Add{func_id}StructHandles(parent_id, id_wrapper->{member.name}->GetPointer(), id_wrapper->{member.name}->GetMetaStructPointer(), handle_struct->{member.name}, object_info_table);\n'
                body += '        }\n'
            elif self.is_struct(member.base_type):
                # This is a struct that includes handles.
                if member.is_array:
                    body += '        AddStructArrayHandles<Decoded_{}>(parent_id, id_wrapper->{name}->GetMetaStructPointer(), id_wrapper->{name}->GetLength(), handle_struct->{name}, static_cast<size_t>(handle_struct->{length}), object_info_table);\n'.format(
                        member.base_type,
                        name=member.name,
                        length=member.array_length
                    )
                elif member.is_pointer:
                    body += '        AddStructArrayHandles<Decoded_{}>(parent_id, id_wrapper->{name}->GetMetaStructPointer(), 1, handle_struct->{name}, 1, object_info_table);\n'.format(
                        member.base_type, name=member.name
                    )
                else:
                    body += '        AddStructHandles(parent_id, id_wrapper->{name}, &handle_struct->{name}, object_info_table);\n'.format(
                        name=member.name
                    )
            else:
                type = member.base_type
                func_id = member.base_type + 'Info'
                type = platform_type + member.base_type[2:] + 'Info'
                object_info_table_add = ', &CommonObjectInfoTable::Add{}'.format(
                    func_id
                )

                # If it is an array or pointer, add with the utility function.
                if (member.is_array or member.is_pointer):
                    if member.is_array:
                        body += '        handle_mapping::AddHandleArray<{}>(parent_id, id_wrapper->{name}.GetPointer(), id_wrapper->{name}.GetLength(), handle_struct->{name}, handle_struct->{length}, object_info_table{});\n'.format(
                            type,
                            object_info_table_add,
                            name=member.name,
                            length=member.array_length
                        )
                    else:
                        body += '        handle_mapping::AddHandleArray<{}>(parent_id, id_wrapper->{name}.GetPointer(), 1, handle_struct->{name}, 1, object_info_table{}});\n'.format(
                            type,
                            object_info_table_add,
                            name=member.name
                        )
                else:
                    body += '        handle_mapping::AddHandle<{}>(parent_id, id_wrapper->{name}, handle_struct->{name}, object_info_table{});\n'.format(
                        type,
                        object_info_table_add,
                        name=member.name
                    )

        body += '    }\n'
        body += '}'
        return body

    def make_struct_handle_allocations(self, name, members):
        """Generate expressions to allocate memory for handles created at replay that are embedded in structs.
        Determine if the struct only contains members that are structs that contain handles or static arrays of handles,
        and does not need a temporary variable referencing the struct value.
        """
        needs_value_ptr = False
        for member in members:
            if self.is_handle(
                member.base_type
            ) and not (member.is_array and not member.is_dynamic):
                needs_value_ptr = True
                break

        body = 'void SetStructHandleLengths(Decoded_{name}* wrapper)\n'.format(
            name=name
        )
        body += '{\n'

        if not needs_value_ptr:
            body += '    if (wrapper != nullptr)\n'
            body += '    {\n'
        else:
            body += '    if ((wrapper != nullptr) && (wrapper->decoded_value != nullptr))\n'
            body += '    {\n'
            body += '        {}* value = wrapper->decoded_value;\n'.format(
                name
            )
            body += '\n'

        for member in members:
            if self.is_extended_struct_definition(member):
                func_id = self.get_extended_struct_func_prefix()
                body += f'        if (wrapper->{member.name})\n'
                body += '        {\n'
                body += f'            Set{func_id}StructHandleLengths(wrapper->{member.name}->GetPointer(), wrapper->{member.name}->GetMetaStructPointer());\n'
                body += '        }\n'
            elif self.is_struct(member.base_type):
                # This is a struct that includes handles.
                if member.is_array:
                    body += '        SetStructArrayHandleLengths<Decoded_{}>(wrapper->{name}->GetMetaStructPointer(), wrapper->{name}->GetLength());\n'.format(
                        member.base_type, name=member.name
                    )
                elif member.is_pointer:
                    body += '        SetStructArrayHandleLengths<Decoded_{}>(wrapper->{name}->GetMetaStructPointer(), 1);\n'.format(
                        member.base_type, name=member.name
                    )
                else:
                    body += '        SetStructHandleLengths(wrapper->{name});\n'.format(
                        name=member.name
                    )
            else:
                # If it is an array or pointer, add with the utility function.
                if (member.is_array or member.is_pointer):
                    if member.is_array:
                        body += '        wrapper->{name}.SetHandleLength(wrapper->{name}.GetLength());\n'.format(
                            name=member.name
                        )
                    else:
                        body += '        wrapper->{}.SetHandleLength(1);\n'.format(
                            member.name
                        )

                    if member.is_dynamic or member.is_pointer:
                        body += '        value->{name} = wrapper->{name}.GetHandlePointer();\n'.format(
                            name=member.name
                        )

        body += '    }\n'
        body += '}'
        return body



#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2020 Valve Corporation
# Copyright (c) 2018-2023 LunarG, Inc.
# Copyright (c) 2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import json
import sys
from openxr_base_generator import OpenXrBaseGenerator, OpenXrBaseGeneratorOptions, write
from khronos_replay_consumer_body_generator import KhronosReplayConsumerBodyGenerator


class OpenXrReplayConsumerBodyGeneratorOptions(OpenXrBaseGeneratorOptions):
    """Options for generating a C++ class for OpenXr capture file replay."""

    def __init__(
        self,
        replay_overrides=None,  # Path to JSON file listing OpenXr API calls to override on replay.
        dump_resources_overrides=None,  # Path to JSON file listing OpenXr API calls to override on replay.
        replay_async_overrides=None,  # Path to JSON file listing OpenXr API calls to override on replay.
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        OpenXrBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            replay_overrides=replay_overrides,
            dump_resources_overrides=dump_resources_overrides,
            replay_async_overrides=replay_async_overrides,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.extend((
            'decode/custom_openxr_struct_handle_mappers.h',
            'decode/custom_vulkan_struct_handle_mappers.h',
            'decode/openxr_handle_mapping_util.h',
            'decode/vulkan_handle_mapping_util.h',
            'generated/generated_openxr_dispatch_table.h',
            'generated/generated_openxr_replay_consumer.h',
            'generated/generated_openxr_struct_handle_mappers.h',
            'util/defines.h',
        ))
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'decode'))

class OpenXrReplayConsumerBodyGenerator(
    KhronosReplayConsumerBodyGenerator, OpenXrBaseGenerator
):
    """OpenXrReplayConsumerBodyGenerator - subclass of OpenXrBaseGenerator.
    Generates C++ member definitions for the OpenXrReplayConsumer class responsible for
    replaying decoded OpenXr API call parameter data.
    Generate a C++ class for OpenXr capture file replay.
    """


    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        OpenXrBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        # These functions should be manual if anything, not code-gened
        # TODO: Check on whether improvements to codegen could make these
        #       possible to codegen
        self.APICALL_BLACKLIST += [
            'xrNegotiateLoaderRuntimeInterface',
            'xrNegotiateLoaderApiLayerInterface',
            'xrCreateInstance',
            'xrCreateApiLayerInstance',
            'xrPollEvent',
            'xrCreateSwapchain',
            'xrReleaseSwapchainImage',
            'xrEndFrame',
            'xrLocateSpaces',
            'xrInitializeLoaderKHR',
            'xrLocateHandJointsEXT',
            'xrGetHandMeshFB',
            'xrLocateBodyJointsFB',
            'xrCreateVulkanInstanceKHR',
            'xrCreateVulkanDeviceKHR',
            'xrCreateDebugUtilsMessengerEXT',
            'xrEnumerateEnvironmentBlendModes',
        ]

        # These structures require a customized manager when they are an output struct
        # in a `Next` chain
        self.OUTPUT_NEXT_OVERRIDES = [
            'XrSpaceVelocities',
            'XrBindingModificationsKHR',
            'XrHandJointVelocitiesEXT',
        ]
        self.NOT_SKIP_FUNCTIONS_OFFSCREEN = []
        self.SKIP_FUNCTIONS_OFFSCREEN = []
        self.types_treated_as_struct = set(['LARGE_INTEGER'])

    def use_instance_table(self, name, typename):
        """OpenXr always uses the instance table"""
        return True

    def determine_handle_to_remove_value(self, command, values):
        if self.needs_remove_handle_expression(command):
            return values[0]
        return None # Must be handled by caller if this isn't a destroy call

    def needs_remove_handle_expression(self, command):
        return command.startswith('xrDestroy')

    def generate_custom_call(self, name, return_type, values):
        # add custom call to replay consumer body

        api_call = 'format::ApiCallId::ApiCall_{}'.format(name)
        custom_update_args = ["this", "call_info"]

        if return_type != 'void':
            custom_update_args.append("returnValue")
        custom_update_args.extend([value.name for value in values])

        if return_type == 'XrResult':
            custom_update_args.append("replay_result")

        call = "    CustomProcess<{}>::UpdateState({});\n".format(
            api_call, ", ".join(custom_update_args)
        )

        return call


    def process_other_khronos_api_handles(self, value, values):
        """ Add handle association """
        process_text = ''
        if self.is_handle(value.base_type) and not value.base_type.startswith('Vk'):
            process_text = '\n    AssociateParent(*out_{}, in_{});'.format(
                    value.name,
                    self.get_parent_id(self.get_api_data(), value, values)
                )

        return process_text

    def treat_as_struct(self, value):
        return value.base_type in self.types_treated_as_struct
    def check_skip_extended_struct_handling(self, struct, struct_type):
        """ OpenXR Base header don't appear on next chains """
        return  'BASE_HEADER' in struct_type

    def generate_custom_extended_struct_handling(self, struct, struct_type):
        """ Method may be overriden.
            None implies no customization
        """

        if struct not in self.OUTPUT_NEXT_OVERRIDES:
            return None

        content = []
        indent = 12*' '
        add = lambda x: content.append(indent + x)
        add(f'case {struct_type}:')
        add('{')
        add(f'    output_struct->next = OverrideOutputStructNext_{struct}(in_next, output_struct);')
        add(f'    break;')
        add('}')
        return '\n'.join(content)

    def endFile(self):
        """Method override."""
        api_data = self.get_api_data()

        KhronosReplayConsumerBodyGenerator.generate_replay_consumer_content(self, api_data)
        KhronosReplayConsumerBodyGenerator.generate_extended_struct_handling(self, api_data)

        self.newline()

        # Finish processing in superclass
        OpenXrBaseGenerator.endFile(self)




#!/usr/bin/env python3
#
# Copyright (c) 2019 Collabora, Ltd.
#
# SPDX-License-Identifier: Apache-2.0
#
# Author(s):    Rylie Pavlik <rylie.pavlik@collabora.com>
#
# Purpose:      This script converts leading comments on some Python
#               classes and functions into docstrings.
#               It doesn't attempt to deal with line continuations, etc.
#               so you may want to "join line" on your def statements
#               temporarily before running.

import re

from spec_tools.file_process import LinewiseFileProcessor

COMMENT_RE = re.compile(r" *#(!.*| (?P<content>.*))?")
CONVERTIBLE_DEF_RE = re.compile(r"(?P<indentation> *)(def|class) .*:")


class CommentConverter(LinewiseFileProcessor):
    def __init__(self, single_line_quotes=False, allow_blank_lines=False):
        super().__init__()
        self.comment_lines = []
        "Temporary storage for contiguous comment lines."

        self.trailing_empty_lines = []
        "Temporary storage for empty lines following a comment."

        self.output_lines = []
        "Fully-processed output lines."

        self.single_line_quotes = single_line_quotes
        "Whether we generate simple, single-line quotes for single line comments."

        self.allow_blank_lines = allow_blank_lines
        "Whether we allow blank lines between a comment and the thing it's considered to document."

        self.done_with_initial_comment = False
        "Have we read our first non-comment line yet?"

    def output_line(self, line=None):
        if line:
            self.output_lines.append(line)
        else:
            self.output_lines.append("")

    def output_normal_line(self, line):
        # flush any comment lines we had stored and output this line.
        self.dump_comment_lines()
        self.output_line(line)

    def dump_comment_lines(self):
        # Early out for empty
        if not self.comment_lines:
            return

        for line in self.comment_lines:
            self.output_line(line)
        self.comment_lines = []

        for line in self.trailing_empty_lines:
            self.output_line(line)
        self.trailing_empty_lines = []

    def dump_converted_comment_lines(self, indent):
        # Early out for empty
        if not self.comment_lines:
            return

        for line in self.trailing_empty_lines:
            self.output_line(line)
        self.trailing_empty_lines = []

        indent = f"{indent}    "

        def extract(line):
            match = COMMENT_RE.match(line)
            content = match.group('content')
            if content:
                return content
            return ""

        # Extract comment content
        lines = [extract(line) for line in self.comment_lines]

        # Drop leading empty comments.
        while lines and not lines[0].strip():
            lines.pop(0)

        # Drop trailing empty comments.
        while lines and not lines[-1].strip():
            lines.pop()

        # Add single- or multi-line-string quote
        if self.single_line_quotes \
            and len(lines) == 1 \
                and '"' not in lines[0]:
            quote = '"'
        else:
            quote = '"""'
        lines[0] = quote + lines[0]
        lines[-1] = lines[-1] + quote

        # Output lines, indenting content as required.
        for line in lines:
            if line:
                self.output_line(indent + line)
            else:
                # Don't indent empty comment lines
                self.output_line()

        # Clear stored comment lines since we processed them
        self.comment_lines = []

    def queue_comment_line(self, line):
        if self.trailing_empty_lines:
            # If we had blank lines between comment lines, they are separate blocks
            self.dump_comment_lines()
        self.comment_lines.append(line)

    def handle_empty_line(self, line):
        """Handle an empty line.

        Contiguous empty lines between a comment and something documentable do not
        disassociate the comment from the documentable thing.
        We have someplace else to store these lines in case there isn't something
        documentable coming up."""
        if self.comment_lines and self.allow_blank_lines:
            self.trailing_empty_lines.append(line)
        else:
            self.output_normal_line(line)

    def is_next_line_doc_comment(self):
        next_line = self.next_line_rstripped
        if next_line is None:
            return False

        return next_line.strip().startswith('"')

    def process_line(self, line_num, line):
        line = line.rstrip()
        comment_match = COMMENT_RE.match(line)
        def_match = CONVERTIBLE_DEF_RE.match(line)

        # First check if this is a comment line.
        if comment_match:
            if self.done_with_initial_comment:
                self.queue_comment_line(line)
            else:
                self.output_line(line)
        else:
            # If not a comment line, then by definition we're done with the comment header.
            self.done_with_initial_comment = True
            if not line.strip():
                self.handle_empty_line(line)
            elif def_match and not self.is_next_line_doc_comment():
                # We got something we can make a docstring for:
                # print the thing the docstring is for first,
                # then the converted comment.

                indent = def_match.group('indentation')
                self.output_line(line)
                self.dump_converted_comment_lines(indent)
            else:
                # Can't make a docstring for this line:
                self.output_normal_line(line)

    def process(self, fn, write=False):
        self.process_file(fn)

        if write:
            with open(fn, 'w', encoding='utf-8') as fp:
                for line in self.output_lines:
                    fp.write(line)
                    fp.write('\n')

        # Reset state
        self.__init__(self.single_line_quotes, self.allow_blank_lines)


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('filenames', metavar='filename',
                        type=str, nargs='+',
                        help='A Python file to transform.')
    parser.add_argument('-b', '--blanklines', action='store_true',
                        help='Allow blank lines between a comment and a define and still convert that comment.')

    args = parser.parse_args()

    converter = CommentConverter(allow_blank_lines=args.blanklines)
    for fn in args.filenames:
        print("Processing", fn)
        converter.process(fn, write=True)


if __name__ == "__main__":
    main()



#!/usr/bin/python3 -i
#
# Copyright (c) 2019 Valve Corporation
# Copyright (c) 2019-2025 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from openxr_base_generator import OpenXrBaseGenerator, OpenXrBaseGeneratorOptions, write
from khronos_dispatch_table_generator import KhronosDispatchTableGenerator

openxr_load_function="""template <typename GetProcAddr, typename Handle, typename FuncP>
static void LoadOpenXrFunction(GetProcAddr gpa, Handle handle, const char* name, FuncP* funcp)
{
    XrResult result = gpa(handle, name, reinterpret_cast<PFN_xrVoidFunction*>(funcp));
    if (result != XR_SUCCESS)
    {
        (*funcp) = nullptr;
    }
}"""

class OpenXrDispatchTableGeneratorOptions(OpenXrBaseGeneratorOptions):
    """Options for generating a dispatch table for OpenXr API calls."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        OpenXrBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.extend((
            'util/defines.h',
            'util/logging.h',
        ))

        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'encode'))

class OpenXrDispatchTableGenerator(OpenXrBaseGenerator, KhronosDispatchTableGenerator):
    """OpenXrDispatchTableGenerator - subclass of OpenXrBaseGenerator.
    Generates a dispatch table for OpenXr API calls.
    Generate dispatch table for OpenXr API calls.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        OpenXrBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )
        # Map of return types to default return values for no-op functions
        self.RETURN_DEFAULTS = {
            'XrResult': 'XR_SUCCESS',
            'XrBool32': 'XR_TRUE',
            'PFN_xrVoidFunction': 'nullptr',
            'uint32_t': '0',
            'uint64_t': '0'
        }

        self.instance_cmd_names = dict(
        )  # Map of API call names to no-op function declarations
        self.device_cmd_names = dict(
        )  # Map of API call names to no-op function declarations


    def is_instance_command(self, api_data, command, first_value):
        """ Method override """
        return self.is_handle(first_value.base_type)

    def write_layer_table_manual_entries(self):
        """ Method must overridden. """
        write(
            '    PFN_xrCreateApiLayerInstance CreateApiLayerInstance{ nullptr };',
            file=self.outFile
        )

    def create_load_function(self, api_data):
        return openxr_load_function

    def endFile(self):
        """Method override."""
        KhronosDispatchTableGenerator.generateDispatchTable(self, gen_dispatch_key=False)

        # Finish processing in superclass
        OpenXrBaseGenerator.endFile(self)



#!/usr/bin/env python3
#
# Copyright (c) 2021 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from dx12_base_generator import Dx12BaseGenerator, write


class Dx12ApiCallEncodersHeaderGenerator(Dx12BaseGenerator):
    """Generates C++ functions responsible for encoding Dx12 API call."""

    # SECURITY_ATTRIBUTES in minwinbase.h has been defined
    # in custom_vulkan_struct_encoders.h
    BLOCK_LIST = ['_SECURITY_ATTRIBUTES']

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )
        self.check_blacklist = True

    def beginFile(self, gen_opts):
        """Method override."""
        Dx12BaseGenerator.beginFile(self, gen_opts)

        self.write_include()

        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(encode)', file=self.outFile)
        self.newline()

    def generate_feature(self):
        """Method override."""
        Dx12BaseGenerator.generate_feature(self)

        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            self.newline()
            write(self.dx12_prefix_strings.format(k), file=self.outFile)
            self.newline()

            for m in v.functions:
                if self.is_required_function_data(m) and (
                    not self.is_cmd_black_listed(m['name'])
                ):
                    self.write_encode_function('', m)

            for class_name, class_value in v.classes.items():
                if self.is_required_class_data(class_value):
                    for m in class_value['methods']['public']:
                        if not self.is_method_black_listed(
                            class_name, m['name']
                        ):
                            self.write_encode_function(class_name, m)

                elif self.is_required_struct_data(class_name, class_value):
                    self.write_encode_struct(
                        class_name, class_value['properties']
                    )

    def write_include(self):
        code = ''
        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            code += '#include <{}>\n'.format(k)

        code += (
            "\n"
            "#include \"encode/parameter_encoder.h\"\n"
            "#include \"generated/generated_dx12_wrappers.h\"\n"
            "#include \"util/defines.h\"\n"
            "\n"
        )

        write(code, file=self.outFile)

    def get_encode_struct_body(self, properties):
        return ';'

    def write_encode_struct(self, name, properties):
        if self.is_block('', name):
            return

        code = (
            'void EncodeStruct(ParameterEncoder* encoder, const {}& value)'.
            format(name)
        )

        code += self.get_encode_struct_body(properties)
        write(code, file=self.outFile)
        self.newline()

    def get_encode_function_body(self, class_name, method_info, is_result):
        return ';'

    def is_block(self, class_name, method_name):
        key = class_name
        if key:
            key += '_'
        key += method_name
        return key in self.BLOCK_LIST

    def write_encode_function(self, class_name, method_info):
        parameters = ''
        if class_name:
            parameters = '    {}_Wrapper* wrapper'.format(class_name)

        is_result = False
        rtn_type = method_info['rtnType']
        if rtn_type.find('void ') == -1 or rtn_type.find('void *') != -1:
            rtn_types1 = self.clean_type_define(rtn_type)

            if class_name:
                parameters += ',\n'
            parameters += '    ' + rtn_types1 + ' return_value'
            if rtn_types1 == 'HRESULT':
                is_result = True

        space_index = 0
        for p in method_info['parameters']:
            if parameters:
                parameters += ',\n'
            parameters += '    '
            parameters += self.clean_type_define(p['type'])
            parameters += ' '
            parameters += p['name']

            if 'array_size' in p:
                array_length = p['array_size']
                parameters += ' '

                if 'multi_dimensional_array' in p:
                    p['multi_dimensional_array']

                    if 'multi_dimensional_array_size' in p:
                        multi_dimensional_array_size = p[
                            'multi_dimensional_array_size']

                        array_sizes = multi_dimensional_array_size.split("x")
                        for size in array_sizes:
                            parameters += '['
                            parameters += size
                            parameters += ']'
                else:
                    parameters += '['
                    parameters += array_length
                    parameters += ']'

            while True:
                space_index = parameters.find(' ', space_index) + 1

                if space_index != 0 and (
                    parameters[space_index] == '*'
                    or parameters[space_index - 2] == '('
                    or parameters[space_index] == '('
                    or parameters[space_index] == ')'
                ):
                    parameters = parameters[:space_index
                                            - 1] + parameters[space_index:]
                elif space_index == 0:
                    break

        _class_name = ''
        if class_name:
            _class_name = '_' + class_name

        code = 'void Encode{}_{}(\n'\
            '{})'.format(_class_name, method_info['name'], parameters)

        code += self.get_encode_function_body(
            class_name, method_info, is_result
        )
        write(code, file=self.outFile)
        self.newline()

    def endFile(self):
        """Method override."""
        self.newline()
        write('GFXRECON_END_NAMESPACE(encode)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        Dx12BaseGenerator.endFile(self)



#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2019 Valve Corporation
# Copyright (c) 2018-2019 LunarG, Inc.
# Copyright (c) 2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from openxr_base_generator import OpenXrBaseGenerator, OpenXrBaseGeneratorOptions, write
from khronos_api_call_encoders_generator import KhronosApiCallEncodersGenerator

class OpenXrApiCallEncodersBodyGeneratorOptions(OpenXrBaseGeneratorOptions):
    """Options for generating C++ functions for OpenXr API parameter encoding."""

    def __init__(
        self,
        capture_overrides=None,  # Path to JSON file listing OpenXr API calls to override on capture.
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        OpenXrBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )
        self.capture_overrides = capture_overrides

        begin_end = self.begin_end_file_data
        begin_end.specific_headers.extend((
            'encode/custom_openxr_encoder_commands.h',
            'encode/custom_openxr_struct_handle_wrappers.h',
            'encode/openxr_capture_manager.h',
            'encode/openxr_handle_wrappers.h',
            'encode/openxr_handle_wrapper_util.h',
            'encode/parameter_encoder.h',
            'encode/struct_pointer_encoder.h',
            '',
            'format/api_call_id.h',
            '',
            'generated/generated_openxr_api_call_encoders.h',
            'generated/generated_openxr_struct_handle_wrappers.h',
            'generated/generated_vulkan_struct_handle_wrappers.h',
            '',
            'util/defines.h',
        ))
        begin_end.namespaces.extend(('gfxrecon', 'encode'))


class OpenXrApiCallEncodersBodyGenerator(OpenXrBaseGenerator, KhronosApiCallEncodersGenerator):
    """OpenXrApiCallEncodersBodyGenerator - subclass of OpenXrBaseGenerator.
    Generates C++ functions responsible for encoding OpenXr API call
    parameter data.
    """

    # Map of OpenXr function names to override function names.  Calls to OpenXr functions in the map
    # will be replaced by the override value.
    CAPTURE_OVERRIDES = {}

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        OpenXrBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )
        KhronosApiCallEncodersGenerator.__init__(self)

    def beginFile(self, gen_opts):
        """Method override."""
        OpenXrBaseGenerator.beginFile(self, gen_opts)

        if gen_opts.capture_overrides:
            self.load_capture_overrides(gen_opts.capture_overrides)

    def endFile(self):
        """Method override."""
        self.write_api_call_encoders_contents(make_cmd_body=self.make_cmd_body)

        # Finish processing in superclass
        OpenXrBaseGenerator.endFile(self)

    def use_instance_table(self, name, typename):
        """Check for dispatchable handle types associated with the instance dispatch table."""
        if typename in ['VkInstance', 'VkPhysicalDevice']:
            return True
        return False

    def make_layer_dispatch_call(self, name, values, arg_list):
        """Generate the layer dispatch call invocation."""
        call_setup_expr = []
        object_name = values[0].name
        dispatchfunc = 'openxr_wrappers::GetInstanceTable'

        return [
            call_setup_expr, '{}({})->{}({})'.format(
                dispatchfunc, object_name, name[2:], arg_list
            )
        ]

    def make_cmd_body(self, return_type, name, values):
        """Command definition."""
        indent = ' ' * self.INDENT_SIZE
        is_override = name in self.CAPTURE_OVERRIDES
        encode_after = False
        omit_output_param = None
        has_outputs = self.has_outputs(return_type, values)
        arg_list = self.make_arg_list(values)

        body = ''

        if has_outputs or (return_type and return_type != 'void'):
            encode_after = True

        omit_output_param = None
        if has_outputs and (return_type and return_type != 'void'):
            omit_output_param = 'omit_output_data'
            body += f'{indent}bool {omit_output_param} = false;\n'
            body += '\n'

        top_indent = indent

        if name == "xrCreateApiLayerInstance":
            capture_manager = 'OpenXrCaptureManager::Get()'
            body += indent + 'auto api_call_lock = OpenXrCaptureManager::AcquireExclusiveApiCallLock();\n'
        else:
            capture_manager = 'manager'
            body += indent + 'OpenXrCaptureManager* manager = OpenXrCaptureManager::Get();\n'
            body += indent + 'GFXRECON_ASSERT(manager != nullptr);\n'
            if not is_override:
                # Allow customization that is unlocked and validly reentrant
                # For example if one needs to record calls into the output stream to add context or state
                # to allow replay time operations, or handle differences in state tracking between API's (e.g. wrapped handles)
                body += indent + 'CustomEncoderPreCall<format::ApiCallId::ApiCall_{}>::PreLockReentrant({}, {});\n'.format(
                    name, capture_manager, arg_list
                )

                # Declare for handles that need unwrapping.
                unwrapped_arg_list, unwrap_list = self.make_handle_unwrapping(
                    name, values
                )
                if unwrap_list:
                    body += indent + f'HandleUnwrapMemory* handle_unwrap_memory = nullptr;\n'
                    body += '\n'.join(
                        [
                            indent +
                            '{type} {wrap_name} = nullptr;\n'.format(**unwrap)
                            for unwrap in unwrap_list
                        ]
                    )

                body += indent + 'CommonCaptureManager::CaptureMode save_capture_mode;\n'
                top_indent = indent + ' ' * self.INDENT_SIZE
                body += indent + '{\n'

                lock_call = 'auto call_lock = manager->AcquireCallLock();\n'
                body += top_indent + lock_call

        body += '\n'

        body += top_indent + 'CustomEncoderPreCall<format::ApiCallId::ApiCall_{}>::Dispatch({}, {});\n'.format(
            name, capture_manager, arg_list
        )

        if not encode_after:
            body += '\n'
            body += self.make_parameter_encoding(
                name, values, return_type, top_indent, omit_output_param
            )

        # Some API calls have different Success criteria for outputting data
        emit_output_check = f'CustomCallResult<format::ApiCallId::ApiCall_{name}>::Succeeded ({capture_manager}, result)'

        if is_override:
            # Capture overrides simply call the override function without handle unwrap/wrap
            # Construct the function call to dispatch to the next layer.
            call_expr = '{}({})'.format(
                self.CAPTURE_OVERRIDES[name], self.make_arg_list(values)
            )
            if return_type and return_type != 'void':
                body += indent + '{} result = {};\n'.format(
                    return_type, call_expr
                )
            else:
                body += indent + '{};\n'.format(call_expr)

            if has_outputs and (return_type and return_type != 'void'):
                body += f'{indent}{omit_output_param} = !{emit_output_check};\n'
        else:
            # Unwrap handles that need unwrapping
            if unwrap_list:
                body += '\n' + top_indent + f'handle_unwrap_memory = {capture_manager}->GetHandleUnwrapMemory();\n'
                body += '\n'.join(
                    [
                        top_indent
                        + '{wrap_name} = {call};\n'.format(**unwrap)
                        for unwrap in unwrap_list
                    ]
                )

            # Disable capture for reentrance
            body += top_indent + 'save_capture_mode = manager->GetCaptureMode();\n'
            body += top_indent + 'manager->SetCaptureMode(CommonCaptureManager::CaptureModeFlags::kModeDisabled);\n'

            # Unlock above (only !is_override)
            body += indent + '}\n\n'

            # Still holding a lock across a call here...
            if self.lock_for_destroy_handle_is_needed(name):
                body += indent + 'ScopedDestroyLock exclusive_scoped_lock;\n'

            # Construct the function call to dispatch to the next layer.
            (call_setup_expr, call_expr) = self.make_layer_dispatch_call(
                name, values, unwrapped_arg_list
            )
            if call_setup_expr:
                for e in call_setup_expr:
                    body += indent + e + '\n'
            if return_type and return_type != 'void':
                body += indent + '{} result = {};\n'.format(
                    return_type, call_expr
                )
            else:
                body += indent + '{};\n'.format(call_expr)

            # Need to relock, since lock was released before dispatch
            body += '\n' + indent + lock_call

            # Restore capture_mode
            body += indent + 'manager->SetCaptureMode(save_capture_mode);\n'

            # Wrap newly created handles.
            wrap_expr = self.make_handle_wrapping(values, indent)
            if wrap_expr:
                body += '\n'
                if return_type and return_type != 'void':
                    body += f'{indent}if ({emit_output_check})\n'
                    body += indent + '{\n'
                    body += '    ' + wrap_expr
                    body += indent + '}\n'
                    if has_outputs:
                        body += f'{indent}else\n'
                        body += f'{indent}{{\n'
                        body += f'{indent}    {omit_output_param} = true;\n'
                        body += f'{indent}}}\n'
                else:
                    body += wrap_expr
            elif has_outputs and (return_type and return_type != 'void'):
                body += f'{indent}{omit_output_param} = !{emit_output_check};\n'

        if encode_after:
            body += self.make_parameter_encoding(
                name, values, return_type, indent, omit_output_param
            )

        body += '\n'
        if return_type and return_type != 'void':
            body += '    CustomEncoderPostCall<format::ApiCallId::ApiCall_{}>::Dispatch({}, result, {});\n'.format(
                name, capture_manager, arg_list
            )
        else:
            body += '    CustomEncoderPostCall<format::ApiCallId::ApiCall_{}>::Dispatch({}, {});\n'.format(
                name, capture_manager, arg_list
            )

        cleanup_expr = self.make_handle_cleanup(name, values, indent)
        if cleanup_expr:
            body += '\n'
            body += cleanup_expr

        if return_type and return_type != 'void':
            body += '\n'
            body += '    return result;\n'

        return body

    def make_begin_api_call(self, name, values):
        capture_manager = 'manager'
        if name == 'xrCreateApiLayerInstance':
            capture_manager = 'OpenXrCaptureManager::Get()'

        if name.startswith('xrCreate') or name.startswith('xrDestroy'):
            return 'auto encoder = {}->BeginTrackedApiCallCapture(format::ApiCallId::ApiCall_{});\n'.format(
                capture_manager, name
            )
        else:
            return 'auto encoder = {}->BeginApiCallCapture(format::ApiCallId::ApiCall_{});\n'.format(
                capture_manager, name
            )

    def make_end_api_call(self, name, values, return_type):
        decl = 'manager->'
        if name == 'xrCreateApiLayerInstance':
            decl = 'OpenXrCaptureManager::Get()->'

        if name.startswith('xrCreate'):
            if name == 'xrCreateSwapchainAndroidSurfaceKHR' or 'Vulkan' in name:
                # The handle is the second last parameter.
                handle = values[-2]
                parent_handle = values[0] if self.is_handle(
                    values[0].base_type
                ) else None
            else:
                # The handle is the last parameter.
                handle = values[-1]
                parent_handle = values[0] if self.is_handle(
                    values[0].base_type
                ) else None

            handle_wrapper = self.get_handle_wrapper(handle.base_type)

            #  Search for the create info struct
            info_base_type = 'void'
            info_name = 'nullptr'
            for value in values:
                if ('CreateInfo' in value.base_type):
                    info_base_type = value.base_type
                    info_name = value.name
                    # Confirm array counts match
                    if value.is_array and (
                        handle.array_length != value.array_length
                    ):
                        print(
                            'WARNING: {} has separate array counts for create info structures ({}) and handles ({})'
                            .format(
                                name, value.array_length, handle.array_length
                            )
                        )

            return_value = 'XR_SUCCESS'
            if return_type == 'XrResult':
                return_value = 'result'

            if handle.is_array:
                length_name = self.make_array_length_expression(handle)

                if 'pAllocateInfo->' in length_name:
                    # This is a pool allocation call, which receives one allocate info structure that is shared by all object being allocated.
                    decl += 'EndPoolCreateApiCallCapture<{}, {}, {}>({}, {}, {}, {}, {})'.format(
                        parent_handle.base_type, handle_wrapper,
                        info_base_type, return_value, parent_handle.name,
                        length_name, handle.name, info_name
                    )
                else:
                    # This is a multi-object creation call (e.g. pipeline creation, or swapchain image retrieval), which receives
                    # separate create info structures for each object being created. Many multi-object creation calls receive a
                    # handle as their second parameter, which is of interest to the state tracker (e.g. the hPipelineCache handle
                    # from vkCreateGraphicsPipelines or the vkSwapchain handle from vkGetSwapchainImagesKHR). For api calls that do
                    # not receive a handle as the second parameter (e.g. vkEnumeratePhysicalDevices), the handle type is set to 'void*'.
                    if handle.base_type in self.struct_names:
                        # "handle" is actually a struct with embedded handles
                        unwrap_handle_def = 'nullptr'
                        member_handle_type, member_handle_name, member_array_length = self.get_struct_handle_member_info(
                            self.structs_with_handles[handle.base_type]
                        )
                        member_handle_wrapper = self.get_wrapper_prefix_from_type(
                            member_handle_type
                        )

                        if not member_array_length:
                            unwrap_handle_def = '[]({}* handle_struct)->{wrapper}r* {{ return openxr_wrappers::GetWrapper<{wrapper}>(handle_struct->{}); }}'.format(
                                handle.base_type,
                                member_handle_name,
                                wrapper=member_handle_wrapper
                            )

                        decl += 'EndStructGroupCreateApiCallCapture<{}, {}, {}>({}, {}, {}, {}, {})'.format(
                            parent_handle.base_type, member_handle_wrapper,
                            handle.base_type, return_value, parent_handle.name,
                            length_name, handle.name, unwrap_handle_def
                        )
                    elif self.is_handle(values[1].base_type):
                        second_handle = values[1]
                        decl += 'EndGroupCreateApiCallCapture<{}, {}, {}, {}>({}, {}, {}, {}, {}, {})'.format(
                            parent_handle.base_type, second_handle.base_type,
                            handle_wrapper, info_base_type, return_value,
                            parent_handle.name, second_handle.name,
                            length_name, handle.name, info_name
                        )
                    else:
                        decl += 'EndGroupCreateApiCallCapture<{}, void*, {}, {}>({}, {}, nullptr, {}, {}, {})'.format(
                            parent_handle.base_type, handle_wrapper,
                            info_base_type, return_value, parent_handle.name,
                            length_name, handle.name, info_name
                        )

            else:
                end_api_call = 'EndCreateApiCallCapture'
                if handle.base_type in self.atom_names:
                    end_api_call = 'EndCreateAtomApiCallCapture'

                if handle.base_type in self.struct_names:
                    # TODO: No cases in current OpenXr spec of handle inside non-array output structure
                    raise NotImplementedError
                elif parent_handle:
                    decl += '{}<{}, {}, {}>({}, {}, {}, {})'.format(
                        end_api_call, parent_handle.base_type, handle_wrapper,
                        info_base_type, return_value, parent_handle.name,
                        handle.name, info_name
                    )
                else:
                    # Instance creation does not have a parent handle; set the parent handle type to 'void*'.
                    decl += '{}<const void*, {}, {}>({}, nullptr, {}, {})'.format(
                        end_api_call, handle_wrapper, info_base_type,
                        return_value, handle.name, info_name
                    )

        elif name.startswith('xrDestroy'):
            handle = values[0]
            handle_wrapper = self.get_handle_wrapper(handle.base_type)

            if handle.is_array:
                decl += 'EndDestroyApiCallCapture<{}>({}, {})'.format(
                    handle_wrapper, handle.array_length, handle.name
                )
            else:
                decl += 'EndDestroyApiCallCapture<{}>({})'.format(
                    handle_wrapper, handle.name
                )
        else:
            decl += 'EndApiCallCapture()'

        decl += ';\n'
        return decl

    def make_handle_wrapping(self, values, indent):
        expr = ''
        for value in values:
            if self.is_output_parameter(value) and (
                self.is_handle_like(value.base_type) or self.struct_has_handles(value.base_type)
            ):
                parent_prefix = 'openxr_wrappers::'
                co_parent_prefix = 'openxr_wrappers::'
                temp_prefix = self.get_wrapper_prefix_from_type(
                    values[0].base_type
                )
                if temp_prefix != 'UNKNOWN_WRAPPERS':
                    parent_prefix = temp_prefix + '::'

                if values[1] is not None:
                    temp_prefix = self.get_wrapper_prefix_from_type(
                        values[1].base_type
                    )
                    if temp_prefix != 'UNKNOWN_WRAPPERS':
                        co_parent_prefix = temp_prefix + '::'

                # The XrInstance handle does not have parent, so the 'unused'
                # values will be provided to the wrapper creation function.
                parent_type = parent_prefix + 'NoParentWrapper'
                parent_value = parent_prefix + 'NoParentWrapper::kHandleValue'
                if self.is_handle_like(values[0].base_type):
                    parent_type = self.get_handle_wrapper(values[0].base_type)
                    parent_value = values[0].name

                # Some handles have two parent handles, such as swapchain images and display modes,
                # or command buffers and descriptor sets allocated from pools.
                co_parent_type = co_parent_prefix + 'NoParentWrapper'
                co_parent_value = co_parent_prefix + 'NoParentWrapper::kHandleValue'
                if self.is_handle_like(
                    values[1].base_type
                ) and not self.is_output_parameter(values[1]):
                    co_parent_type = self.get_handle_wrapper(
                        values[1].base_type
                    )
                    co_parent_value = values[1].name

                wrapper_prefix = self.get_wrapper_prefix_from_type(
                    value.base_type
                )

                if value.is_array:
                    length_name = value.array_length
                    for len in values:
                        if (len.name == length_name) and len.is_pointer:
                            length_name = '({name} != nullptr) ? (*{name}) : 0'.format(
                                name=length_name
                            )
                            break
                    if self.is_handle(value.base_type):

                        handle_wrapper = self.get_handle_wrapper(
                            value.base_type
                        )
                        expr += indent + '{}CreateWrappedHandles<{}, {}, {}>({}, {}, {}, {}, OpenXrCaptureManager::GetUniqueId);\n'.format(
                            parent_prefix, parent_type, co_parent_type,
                            handle_wrapper, parent_value, co_parent_value,
                            value.name, length_name
                        )
                    elif self.is_atom(value.base_type):

                        handle_wrapper = self.get_handle_wrapper(
                            value.base_type
                        )
                        expr += indent + '{}CreateWrappedAtoms<{}, {}>({}, {}, {}, OpenXrCaptureManager::GetUniqueId);\n'.format(
                            parent_prefix, parent_type, handle_wrapper,
                            parent_value, value.name, length_name
                        )
                    elif self.is_opaque(value.base_type):

                        handle_wrapper = self.get_handle_wrapper(
                            value.base_type
                        )
                        expr += indent + '{}CreateWrappedOpaques<{}, {}>({}, {}, {}, OpenXrCaptureManager::GetUniqueId);\n'.format(
                            parent_prefix, parent_type, handle_wrapper,
                            parent_value, value.name, length_name
                        )
                    elif self.is_struct(
                        value.base_type
                    ) and (value.base_type in self.structs_with_handles):
                        expr += indent + '{}CreateWrappedStructArrayHandles<{}, {}, {}>({}, {}, {}, {}, OpenXrCaptureManager::GetUniqueId);\n'.format(
                            parent_prefix, parent_type, co_parent_type,
                            value.base_type, parent_value, co_parent_value,
                            value.name, length_name
                        )
                else:
                    is_handle = self.is_handle(value.base_type)
                    is_atom = self.is_atom(value.base_type)
                    is_opaque = self.is_opaque(value.base_type)
                    if self.is_handle(value.base_type):
                        handle_wrapper = self.get_handle_wrapper(
                            value.base_type
                        )
                        expr += indent + '{}CreateWrappedHandle<{}, {}, {}>({}, {}, {}, OpenXrCaptureManager::GetUniqueId);\n'.format(
                            parent_prefix, parent_type, co_parent_type,
                            handle_wrapper, parent_value, co_parent_value,
                            value.name
                        )
                    elif self.is_atom(value.base_type):
                        handle_wrapper = self.get_handle_wrapper(
                            value.base_type
                        )
                        expr += indent + '{}CreateWrappedAtom<{}, {}>({}, {}, OpenXrCaptureManager::GetUniqueId);\n'.format(
                            parent_prefix, parent_type, handle_wrapper,
                            parent_value, value.name
                        )
                    elif self.is_opaque(value.base_type):
                        handle_wrapper = self.get_handle_wrapper(
                            value.base_type
                        )
                        expr += indent + '{}CreateWrappedOpaque<{}, {}>({}, {}, OpenXrCaptureManager::GetUniqueId);\n'.format(
                            parent_prefix, parent_type, handle_wrapper,
                            parent_value, value.name
                        )
                    elif self.is_struct(
                        value.base_type
                    ) and (value.base_type in self.structs_with_handles):
                        expr += indent + '{}CreateWrappedStructHandles<{}, {}>({}, {}, {}, OpenXrCaptureManager::GetUniqueId);\n'.format(
                            parent_prefix, parent_type, co_parent_type,
                            parent_value, co_parent_value, value.name
                        )
        return expr

    def make_handle_unwrapping(self, name, values):
        args = []
        unwraps = []
        need_unwrap_memory = False
        for value in values:
            arg_name = value.name
            if value.is_pointer or value.is_array:
                if self.is_input_pointer(value):
                    if (value.base_type in self.structs_with_handles
                        ) or (value.base_type in self.GENERIC_HANDLE_STRUCTS):
                        arg_name += '_unwrapped'
                        wrapper_prefix = self.get_wrapper_prefix_from_type(
                            value.base_type
                        ) + '::'
                        unwrap = {
                            'type': value.full_type,
                            'name': value.name,
                            'wrap_name': arg_name,
                            'prefix': wrapper_prefix
                        }
                        if value.is_array:
                            unwrap['length'] = value.array_length
                            unwrap[
                                'call'
                            ] = '{prefix}UnwrapStructArrayHandles({name}, {length}, handle_unwrap_memory)'.format(
                                **unwrap
                            )
                        else:
                            unwrap[
                                'call'
                            ] = '{prefix}UnwrapStructPtrHandles({name}, handle_unwrap_memory)'.format(
                                **unwrap
                            )
                        unwraps.append(unwrap)
            args.append(arg_name)
        return ', '.join(args), unwraps

    def lock_for_destroy_handle_is_needed(self, name):
        if name.startswith('xrDestroy'):
            return True
        else:
            return False

    def make_handle_cleanup(self, name, values, indent):
        expr = ''
        if name.startswith('xrDestroy'):
            handle = values[0]
            wrapper_prefix = self.get_wrapper_prefix_from_type(
                handle.base_type
            ) + '::'
            handle_prefix = self.get_handle_wrapper(handle.base_type)

            if handle.is_array:
                expr += indent + '{}DestroyWrappedHandles<{}>({}, {});\n'.format(
                    wrapper_prefix, handle_prefix, handle.name,
                    handle.array_length
                )
            else:
                expr += indent + '{}DestroyWrappedHandle<{}>({});\n'.format(
                    wrapper_prefix, handle_prefix, handle.name
                )
        return expr



#!/usr/bin/env python3

# Copyright (c) 2019 Advanced Micro Devices, Inc. All rights reserved
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

'''
GFXReconstruct check code style script
'''

import subprocess
import sys
import argparse
import os

def check_code_style(file_list, compare_base, style_script=None, style_config_dir=None):
    """
    Run clang format diff on the changed source file(s)
    against current branch. If branch retrieved failed, 
    fall back to diff against master branch.
    :raises Exception: when formatting errors found
    """
    git_branch = subprocess.check_output(["git", "rev-parse", "--abbrev-ref", compare_base]).strip()
    if git_branch == b'':
        git_branch = "master"
    if not style_script:
        # Look for the clang-format-diff.py script in the same directory as the calling script
        style_script = os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])), "clang-format-diff.py")
    if not style_config_dir:
        # Use the script directory's parent directory as the .clang-format file location
        style_config_dir = os.path.join(os.path.dirname(style_script), os.pardir)
    diff_cmd = ["git", "diff", "-U0", str(git_branch, 'utf-8'), "--"] + file_list
    style_cmd = [sys.executable, style_script,"-p1","-style=file"]
    try:
        diff_process = subprocess.Popen(diff_cmd, stdout=subprocess.PIPE)
        style_output = subprocess.check_output(style_cmd, stdin=diff_process.stdout, cwd=style_config_dir)

        if style_output != b'':
            raise Exception("FORMATTING ERROR!\n" + str(style_output, 'utf-8'))
    except subprocess.CalledProcessError as e:
        print("ERROR calling clang-format-diff.py [{}]".format(e))
        sys.exit(1)

# Main entry point
if '__main__' == __name__:
    # parse params and run check code style function
    parser = argparse.ArgumentParser(prog='check_code_style',
                                     description='Checks if file match the code style specification')
    parser.add_argument('--sourcefile', nargs='+', dest='file_list', help="The source file(s)")
    parser.add_argument('--base', dest='compare_base', metavar='BASE', action='store', default='HEAD',
        help='Git branch name or commit ID to use as the base for C++ code style comparison')
    args = parser.parse_args()
    check_code_style(args.file_list, args.compare_base)
    sys.exit(0)



#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2020 Valve Corporation
# Copyright (c) 2018-2020 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from dx12_base_generator import ValueInfo, write


class Dx12BaseDecoderBodyGenerator():
    """Base class for generating decoder body code."""

    def generate_feature(self):
        """Performs C++ code generation for the feature."""
        platform_type = self.get_api_prefix()

        first = True
        for cmd in self.get_filtered_cmd_names():
            self.cmd_names.append(cmd)

            info = self.feature_cmd_params[cmd]
            return_type = info[0]
            values = info[2]

            cmddef = '' if first else '\n'
            cmddef += 'size_t {}Decoder::Decode_{}(const ApiCallInfo& call_info, const uint8_t* parameter_buffer, size_t buffer_size)\n'.format(
                platform_type, cmd
            )
            cmddef += '{\n'
            cmddef += '    size_t bytes_read = 0;\n'
            cmddef += '\n'
            cmddef += self.make_cmd_body(return_type, cmd, values)
            cmddef += '\n'
            cmddef += '    return bytes_read;\n'
            cmddef += '}'

            write(cmddef, file=self.outFile)
            first = False

    def make_cmd_body(self, return_type, name, values, dx12_method=False):
        """Generate C++ code for the decoder method body."""
        body = ''
        arg_names = []

        # Declarations for decoded types.
        for value in values:
            decode_type = self.make_decoded_param_type(value)
            body += '    {} {};\n'.format(decode_type, value.name)

            if decode_type == 'Decoded_{}'.format(value.base_type):
                body += '    {} value_{};\n'.format(
                    value.base_type, value.name
                )
                body += '    {0}.decoded_value = &value_{0};\n'.format(
                    value.name
                )

            if 'Decoder' in decode_type:
                arg_names.append('&{}'.format(value.name))
            else:
                arg_names.append(value.name)

        # Vulkan return is very simple. Value is only for Dx12 Method.
        dx12_return_value = None
        dx12_return_decode_type = None
        if return_type and return_type != 'void':
            if dx12_method:
                dx12_return_value = self.get_return_value_info(
                    return_type, name
                )
                dx12_return_decode_type = self.make_decoded_param_type(
                    dx12_return_value
                )
                body += '    {} return_value;\n'.format(
                    dx12_return_decode_type
                )

                if dx12_return_decode_type == 'Decoded_{}'.format(return_type):
                    body += '    {} value_returned;\n'.format(return_type)
                    body += '    return_value.decoded_value = &value_returned;\n'

            else:
                body += '    {} return_value;\n'.format(return_type)

        # Blank line after declarations.
        if values or return_type:
            body += '\n'

        # Decode() method calls for pointer decoder wrappers.
        for value in values:
            body += Dx12BaseDecoderBodyGenerator.make_decode_invocation(
                self, value
            )
        if return_type and return_type != 'void':
            if dx12_method:
                body += Dx12BaseDecoderBodyGenerator.make_decode_invocation(
                    self, dx12_return_value
                )
            else:
                body += Dx12BaseDecoderBodyGenerator.make_decode_invocation(
                    self, ValueInfo('return_value', return_type, return_type)
                )

        # Blank line after Decode() method invocations.
        if values or return_type:
            body += '\n'

        # Make the argument list for the API call
        arglist = ', '.join([arg_name for arg_name in arg_names])
        if return_type and return_type != 'void':
            if dx12_method and dx12_return_decode_type.find('Decoder') != -1:
                arglist = ', '.join(['&return_value', arglist])
            else:
                arglist = ', '.join(['return_value', arglist])

        if dx12_method:
            arglist = 'object_id, ' + arglist

        if arglist[-2:] == ', ':
            arglist = arglist[:-2]
        arglist = 'call_info, ' + arglist

        body += '    for (auto consumer : GetConsumers())\n'
        body += '    {\n'
        body += '        consumer->Process_{}({});\n'.format(name, arglist)
        body += '    }\n'

        return body

    def make_decode_invocation(self, value):
        """Generate parameter decode function/method invocation."""
        buffer_args = '(parameter_buffer + bytes_read), (buffer_size - bytes_read)'
        body = ''

        is_struct = False
        is_class = False
        is_string = False
        is_funcp = False
        is_handle = False

        type_name = self.make_invocation_type_name(value.base_type)

        if self.is_struct(type_name):
            is_struct = True
        elif self.is_class(value):
            is_class = True
        elif type_name in ['String', 'WString']:
            is_string = True
        elif type_name == 'FunctionPtr':
            is_funcp = True
        elif self.is_handle(value.base_type):
            is_handle = True

        # is_pointer will be False for static arrays.
        if value.is_pointer or value.is_array:
            if not is_class and type_name in self.EXTERNAL_OBJECT_TYPES and not value.is_array:
                if value.pointer_count > 1:
                    # Pointer to a pointer to an unknown object type (void**), encoded as a pointer to a 64-bit integer ID.
                    body += '    bytes_read += {}.DecodeVoidPtr({});\n'.format(
                        value.name, buffer_args
                    )
                else:
                    # Pointer to an unknown object type, encoded as a 64-bit integer ID.
                    body += '    bytes_read += ValueDecoder::DecodeAddress({}, &{});\n'.format(
                        buffer_args, value.name
                    )
            else:
                if is_struct or is_string or is_handle or (
                    is_class and value.pointer_count > 1
                ):
                    body += '    bytes_read += {}.Decode({});\n'.format(
                        value.name, buffer_args
                    )
                elif is_class and value.pointer_count == 1:
                    body += '    bytes_read += ValueDecoder::DecodeHandleIdValue({}, &{});\n'.format(
                        buffer_args, value.name
                    )
                elif self.has_basetype(value.base_type):
                    base_type = self.get_basetype(value.base_type)
                    body += '    bytes_read += {}.Decode{}({});\n'.format(
                        value.name, self.encode_types[base_type], buffer_args
                    )
                else:
                    body += '    bytes_read += {}.Decode{}({});\n'.format(
                        value.name, type_name, buffer_args
                    )
        else:
            if is_struct:
                body += '    bytes_read += DecodeStruct({}, &{});\n'.format(
                    buffer_args, value.name
                )
            elif is_funcp:
                body += '    bytes_read += ValueDecoder::DecodeAddress({}, &{});\n'.format(
                    buffer_args, value.name
                )
            elif is_handle:
                body += '    bytes_read += ValueDecoder::DecodeHandleIdValue({}, &{});\n'.format(
                    buffer_args, value.name
                )
            elif self.has_basetype(type_name) :
                base_type = self.get_basetype(type_name)
                body += '    bytes_read += ValueDecoder::Decode{}Value({}, &{});\n'.format(
                    self.encode_types[base_type], buffer_args, value.name
                )
            else:
                body += '    bytes_read += ValueDecoder::Decode{}Value({}, &{});\n'.format(
                    type_name, buffer_args, value.name
                )

        return body

    def generate_decode_cases(self):
        prefix = self.get_api_prefix()
        """Generate the (Platform)Decoder::DecodeFunctionCall method."""

        body = f'void {prefix}Decoder::DecodeFunctionCall(format::ApiCallId             call_id,\n'
        body += '                                       const ApiCallInfo&            call_info,\n'
        body += '                                       const uint8_t*                parameter_buffer,\n'
        body += '                                       size_t                        buffer_size)\n'
        body += '{\n'
        body += '    switch(call_id)\n'
        body += '    {\n'
        body += '    default:\n'
        body += f'        {prefix}DecoderBase::DecodeFunctionCall(call_id, call_info, parameter_buffer, buffer_size);\n'
        body += '        break;\n'
        write(body, file=self.outFile)

        for cmd in self.cmd_names:
            cmddef = '    case format::ApiCallId::ApiCall_{}:\n'.format(cmd)
            cmddef += '        Decode_{}(call_info, parameter_buffer, buffer_size);\n'.format(
                cmd
            )
            cmddef += '        break;'
            write(cmddef, file=self.outFile)

        write('    }', file=self.outFile)
        write('}\n', file=self.outFile)



#!/usr/bin/python3 -i
#
# Copyright (c) 2018 Valve Corporation
# Copyright (c) 2018 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from vulkan_base_generator import VulkanBaseGenerator, VulkanBaseGeneratorOptions, write
from khronos_struct_decoders_forward_generator import KhronosStructDecodersForwardGenerator


class VulkanStructDecodersForwardGeneratorOptions(VulkanBaseGeneratorOptions):
    """Options for generating C++ function and forward type declarations for Vulkan struct decoding."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        VulkanBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.append('util/defines.h')
        self.begin_end_file_data.system_headers.append('cstdint')
        self.begin_end_file_data.namespaces.extend((
            'gfxrecon',
            'decode',
        ))


class VulkanStructDecodersForwardGenerator(VulkanBaseGenerator, KhronosStructDecodersForwardGenerator):
    """VulkanStructDecodersForwardGenerator - subclass of VulkanBaseGenerator.
    Generates C++ type and function declarations for decoding Vulkan API structures.
    Generate C++ function and forward type declarations for Vulkan struct decoding.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        VulkanBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    def endFile(self):
        """Method override."""
        KhronosStructDecodersForwardGenerator.write_struct_decoder_forward_prototypes(self)

        self.newline()

        # Finish processing in superclass
        VulkanBaseGenerator.endFile(self)

    def need_feature_generation(self):
        """Indicates that the current feature has C++ code to generate."""
        if self.feature_struct_members:
            return True
        return False



#!/usr/bin/python3 -i
#
# Copyright (c) 2020 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from vulkan_base_generator import VulkanBaseGenerator, VulkanBaseGeneratorOptions, ValueInfo, write


class VulkanReferencedResourceBodyGeneratorOptions(VulkanBaseGeneratorOptions):
    """Options for generating a C++ class for detecting unreferenced resource handles in a capture file."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        VulkanBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.append('generated/generated_vulkan_referenced_resource_consumer.h')
        self.begin_end_file_data.system_headers.append('cassert')
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'decode'))
        self.begin_end_file_data.common_api_headers = []


class VulkanReferencedResourceBodyGenerator(VulkanBaseGenerator):
    """VulkanReferencedResourceBodyGenerator - subclass of VulkanBaseGenerator.
    Generates C++ member definitions for the VulkanReferencedResource class responsible for
    determining which resource handles are used or unused in a capture file.
    Generate a C++ class for detecting unreferenced resource handles in a capture file.
    """

    # All resource and resource associated handle types to be processed.
    RESOURCE_HANDLE_TYPES = [
        'VkBuffer', 'VkImage', 'VkBufferView', 'VkImageView', 'VkFramebuffer',
        'VkDescriptorSet', 'VkCommandBuffer', 'VkAccelerationStructureKHR'
    ]

    # Handle types that contain resource and child resource handle types.
    CONTAINER_HANDLE_TYPES = ['VkDescriptorSet']

    # Handle types that use resource and child resource handle types.
    USER_HANDLE_TYPES = ['VkCommandBuffer']

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        VulkanBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )
        self.restrict_handles = True  # Determines if the 'is_handle' override limits the handle test to only the values conained by RESOURCE_HANDLE_TYPES.

    def endFile(self):
        """Method override."""
        for cmd, info in self.all_cmd_params.items():
            if self.is_cmd_black_listed(cmd):
                continue

            return_type = info[0]
            params = info[2]

            if params and params[0].base_type == 'VkCommandBuffer':

                # Check for parameters with resource handle types.
                handles = self.get_param_list_handles(params[1:])

                if (handles):
                    # Generate a function to add handles to the command buffer's referenced handle list.
                    cmddef = '\n'

                    # Temporarily remove resource only matching restriction from is_handle() when generating the function signature.
                    self.restrict_handles = False
                    cmddef += self.make_consumer_func_decl(
                        return_type,
                        'VulkanReferencedResourceConsumer::Process_' + cmd,
                        params
                    ) + '\n'
                    self.restrict_handles = True

                    cmddef += '{\n'
                    indent = self.INDENT_SIZE * ' '

                    # Add unreferenced parameter macros.
                    unref_count = 0
                    for param in params[1:]:
                        if param not in handles:
                            cmddef += indent + 'GFXRECON_UNREFERENCED_PARAMETER({});\n'.format(
                                param.name
                            )
                            unref_count += 1
                    if unref_count > 0:
                        cmddef += '\n'

                    for index, handle in enumerate(handles):
                        cmddef += self.track_command_handle(
                            index, params[0].name, handle, indent=indent
                        )
                    cmddef += '}'

                    write(cmddef, file=self.outFile)

        self.newline()

        # Finish processing in superclass
        VulkanBaseGenerator.endFile(self)

    def need_feature_generation(self):
        """Indicates that the current feature has C++ code to generate."""
        if self.feature_cmd_params:
            return True
        return False

    def is_handle(self, base_type):
        """Override method to check for handle type, only matching resource handle types."""
        if self.restrict_handles:
            if base_type in self.RESOURCE_HANDLE_TYPES:
                return True
            return False
        else:
            return VulkanBaseGenerator.is_handle(self, base_type)


    def track_command_handle(
        self, index, command_param_name, value, value_prefix='', indent=''
    ):
        body = ''
        tail = ''
        index_name = None
        count_name = None
        value_name = value_prefix + value.name
        is_handle = self.is_handle(value.base_type)

        if (
            value.is_pointer or value.is_array
        ) and value.name != 'ext_struct_info':
            if index > 0:
                body += '\n'

            access_operator = '->'
            if not value_prefix:
                # If there is no prefix, this is the pointer parameter received by the function, which should never be null.
                body += indent + 'assert({} != nullptr);\n'.format(value.name)
                body += '\n'
                # Add IsNull and HasData checks for the pointer decoder, before accessing its data.
                # Note that this does not handle the decoded struct member cases for static arrays, which would need to use '.' instead of '->'.
                body += indent + 'if (!{prefix}{name}{op}IsNull() && ({prefix}{name}{op}HasData()))\n'.format(
                    prefix=value_prefix, name=value.name, op=access_operator
                )
                body += indent + '{\n'
                tail = indent + '}\n' + tail
                indent += ' ' * self.INDENT_SIZE
            else:
                # If there is a prefix, this is a struct member.  We need to determine the type of access operator to use
                # for the member of a 'decoded' struct type, where handle member types will be HandlePointerDecoder, but
                # struct member types will be unique_ptr<StructPointerDecoder>.
                if is_handle:
                    access_operator = '.'

                body += indent + 'if (!{prefix}{name}{op}IsNull() && ({prefix}{name}{op}HasData()))\n'.format(
                    prefix=value_prefix, name=value.name, op=access_operator
                )

                body += indent + '{\n'
                tail = indent + '}\n' + tail
                indent += ' ' * self.INDENT_SIZE

            # Get the pointer from the pointer decoder object.
            value_name = '{}_ptr'.format(value.name)
            if is_handle:
                body += indent + 'auto {} = {}{}{}GetPointer();\n'.format(
                    value_name, value_prefix, value.name, access_operator
                )
            else:
                body += indent + 'auto {} = {}{}{}GetMetaStructPointer();\n'.format(
                    value_name, value_prefix, value.name, access_operator
                )

            # Add a for loop for an array of values.
            if value.is_array:
                index_name = '{}_index'.format(value.name)
                count_name = '{}_count'.format(value.name)
                body += indent + 'size_t {} = {}{}{}GetLength();\n'.format(
                    count_name, value_prefix, value.name, access_operator
                )
                body += indent + 'for (size_t {i} = 0; {i} < {}; ++{i})\n'.format(
                    count_name, i=index_name
                )

                body += indent + '{\n'
                tail = indent + '}\n' + tail
                indent += ' ' * self.INDENT_SIZE

        # Insert commands to add handles to a container, or to process struct members that contain handles.
        if is_handle:
            if value.is_array:
                value_name = '{}[{}]'.format(value_name, index_name)
            elif value.is_pointer:
                value_name = '(*{})'.format(value_name)

            if value.base_type in self.CONTAINER_HANDLE_TYPES:
                body += indent + 'GetTable().AddContainerToUser({}, {});\n'.format(
                    command_param_name, value_name
                )
            elif value.base_type in self.USER_HANDLE_TYPES:
                body += indent + 'GetTable().AddUserToUser({}, {});\n'.format(
                    command_param_name, value_name
                )
            else:
                body += indent + 'GetTable().AddResourceToUser({}, {});\n'.format(
                    command_param_name, value_name
                )

        elif self.is_struct(
            value.base_type
        ) and (value.base_type in self.structs_with_handles):
            if value.is_array:
                access_operator = '[{}].'.format(index_name)
            else:
                access_operator = '->'

            for index, entry in enumerate(
                self.structs_with_handles[value.base_type]
            ):
                if entry.name == 'pNext':
                    ext_structs_with_handles = [
                        ext_struct for ext_struct in
                        self.registry.validextensionstructs[value.base_type]
                        if ext_struct in self.structs_with_handles
                    ]
                    if ext_structs_with_handles:
                        for ext_struct in ext_structs_with_handles:
                            body += indent + '{\n'
                            indent += ' ' * self.INDENT_SIZE
                            body += indent + 'const auto* ext_struct_info = GetPNextMetaStruct<Decoded_{}>({}->pNext);\n'.format(
                                 ext_struct, value_name
                             )

                            body += indent + 'if (ext_struct_info != nullptr)\n'
                            body += indent + '{\n'
                            indent += ' ' * self.INDENT_SIZE
                            body += self.track_command_handle(
                                index,
                                command_param_name,
                                ValueInfo(
                                    'ext_struct_info', ext_struct,
                                    'const {} *'.format(ext_struct), 1
                                ),
                                '',
                                indent=indent
                            )
                            indent = indent[:-self.INDENT_SIZE]
                            body += indent + '}\n'

                            indent = indent[:-self.INDENT_SIZE]
                            body += indent + '}\n'
                else:
                    body += self.track_command_handle(
                        index, command_param_name, entry,
                        value_name + access_operator, indent
                    )

        return body + tail



#!/usr/bin/python3 -i
#
# Copyright (c) 2021-2025 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from khronos_base_generator import write

# Generates declarations for functions for Khronos state table
class KhronosObjectInfoTableHeaderGenerator():

    # yapf: disable
    def generate_info_table(self, api_data, class_name, base_class):
        add_code = ''
        remove_code = ''
        const_get_code = ''
        get_code = ''
        visit_code = ''
        map_code = ''
        info = {}
        info['api'] = api_data.api_class_prefix
        info['type'] = api_data.type_prefix

        handle_like = set(self.handle_names).union(self.atom_names).union(self.opaque_names)
        for handle_name in sorted(handle_like):
            if handle_name in self.handle_aliases:
                continue
            if not handle_name.startswith(api_data.type_prefix):
                continue

            short_handle_name = handle_name[2:]
            info['handle'] = short_handle_name + 'Info'
            info['function'] = handle_name + 'Info'
            info['map'] = short_handle_name[0].lower() + short_handle_name[1:] + '_map_'
            add_code += '    void Add{function}({api}{handle}&& info) {{ Add{type}ObjectInfo(std::move(info), &{map}); }}\n'.format(**info)
            remove_code += '    void Remove{function}(format::HandleId id) {{ {map}.erase(id); }}\n'.format(**info)
            const_get_code += '    const {api}{handle}* Get{function}(format::HandleId id) const {{ return Get{type}ObjectInfo<{api}{handle}>(id, &{map}); }}\n'.format(**info)
            get_code += '    {api}{handle}* Get{function}(format::HandleId id) {{ return Get{type}ObjectInfo<{api}{handle}>(id, &{map}); }}\n'.format(**info)
            visit_code += '    void Visit{function}(std::function<void(const {api}{handle}*)> visitor) const {{  for (const auto& entry : {map}) {{ visitor(&entry.second); }}  }}\n'.format(**info)
            map_code += '     std::unordered_map<format::HandleId, {api}{handle}> {map};\n'.format(**info)

        code = '\n'
        code += f'class {class_name} : {base_class}\n'
        code += '{\n'
        code += '  public:\n'
        code += f'    {class_name}() {{}}\n'
        code += f'    ~{class_name}() {{}}\n'
        code += '\n'
        code += add_code
        code += '\n'
        code += remove_code
        code += '\n'
        code += const_get_code
        code += '\n'
        code += get_code
        code += '\n'
        code += visit_code
        code += '\n'
        code += '  protected:\n'
        code += map_code
        code += '};\n'
        write(code, file=self.outFile)
    # yapf: enable


#!/usr/bin/python3 -i
#
# Copyright (c) 2018 Valve Corporation
# Copyright (c) 2018 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from khronos_base_generator import write


class KhronosEncodeExtendedStructGenerator():
    """KhronosEncodeExtendedStructGenerator.
    Generates C++ code for a Khronos API's extended structure encoding.
    """

    @staticmethod
    def update_begin_end_file_data(lower_api_name, begin_end):
        # Get the current API and generate the items relavent to that
        begin_end.specific_headers.extend((
            'generated/generated_{}_struct_encoders.h'.format(lower_api_name),
            '',
            'encode/parameter_encoder.h',
            'encode/struct_pointer_encoder.h',
            'encode/{}_capture_manager.h'.format(lower_api_name),
            'util/defines.h',
        ))

        begin_end.system_headers.extend((
            'cassert',
            'cstdio',
            'memory',
        ))

        begin_end.namespaces.extend(('gfxrecon', 'encode'))

    def get_encode_struct_while_loop_statement(self, current_api_data):
        """
        Method intended to be overridden.
        This is because some of the loop determination is API-specific
        or may not be needed at all
        """
        return ''


    def write_encode_struct_definition_prefix(self):
        # Get the current API and generate the items relavent to that
        current_api_data = self.get_api_data()

        write(
            'void Encode{}Struct(ParameterEncoder* encoder, const void* value)'
            .format(current_api_data.extended_struct_func_prefix),
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    assert(encoder != nullptr);', file=self.outFile)
        self.newline()
        write(
            '    auto base = reinterpret_cast<const {}*>(value);'.format(
                current_api_data.base_in_struct
            ),
            file=self.outFile
        )
        self.newline()

        loop_statement = self.get_encode_struct_while_loop_statement(current_api_data)
        if loop_statement:
            write(
                '    // Ignore the structures added to the {} chain by the loader.'
                .format(current_api_data.extended_struct_variable),
                file=self.outFile
            )
            write(loop_statement, file=self.outFile)
            write('    {', file=self.outFile)
            write(
                '        base = base->{};'.format(
                    current_api_data.extended_struct_variable
                ),
                file=self.outFile
            )
            write('    }', file=self.outFile)
            self.newline()

        write('    if (base != nullptr)', file=self.outFile)
        write('    {', file=self.outFile)
        write(
            '        switch (base->{})'.format(
                current_api_data.struct_type_variable
            ),
            file=self.outFile
        )
        write('        {', file=self.outFile)
        write('        default:', file=self.outFile)
        write('            {', file=self.outFile)
        write(
            '                // {} is unrecognized.  Write warning message to indicate it will be omitted from the capture and check to see if it points to a recognized value.'
            .format(current_api_data.extended_struct_variable),
            file=self.outFile
        )
        write(
            '                int32_t message_size = std::snprintf(nullptr, 0, "A {} value with unrecognized {} = %d was omitted from the capture file, which may cause replay to fail.", base->{});'
            .format(
                current_api_data.extended_struct_variable,
                current_api_data.struct_type_enum,
                current_api_data.struct_type_variable
            ),
            file=self.outFile
        )
        write(
            '                std::unique_ptr<char[]> message = std::make_unique<char[]>(message_size + 1); // Add 1 for null-terminator.',
            file=self.outFile
        )
        write(
            '                std::snprintf(message.get(), (message_size + 1), "A {} value with unrecognized {} = %d was omitted from the capture file, which may cause replay to fail.", base->{});'
            .format(
                current_api_data.extended_struct_variable,
                current_api_data.struct_type_enum,
                current_api_data.struct_type_variable
            ),
            file=self.outFile
        )
        write(
            '                {}CaptureManager::Get()->WriteDisplayMessageCmd(message.get());'
            .format(current_api_data.api_class_prefix),
            file=self.outFile
        )
        write(
            '                GFXRECON_LOG_WARNING("%s", message.get());',
            file=self.outFile
        )
        write(
            '                Encode{}Struct(encoder, base->{});'.format(
                current_api_data.extended_struct_func_prefix,
                current_api_data.extended_struct_variable
            ),
            file=self.outFile
        )
        write('            }', file=self.outFile)
        write('            break;', file=self.outFile)

    def write_encode_struct_definition_data(self):
        current_api_data = self.get_api_data()

        extended_list = []
        for struct in self.all_extended_structs:
            for ext_struct in self.all_extended_structs[struct]:
                if ext_struct not in extended_list and ext_struct not in self.all_struct_aliases:
                    extended_list.append(ext_struct)

        for struct in sorted(extended_list):
            if struct not in self.struct_type_names:
                continue

            stype = self.struct_type_names[struct]
            write('        case {}:'.format(stype), file=self.outFile)
            write(
                '            EncodeStructPtr(encoder, reinterpret_cast<const {}*>(base));'
                .format(struct),
                file=self.outFile
            )
            write('            break;', file=self.outFile)

        current_api_data = self.get_api_data()

        write('        }', file=self.outFile)
        write('    }', file=self.outFile)
        write('    else', file=self.outFile)
        write('    {', file=self.outFile)
        write(
            '        // {} was either NULL or an ignored loader specific struct.  Write an encoding for a NULL pointer.'
            .format(current_api_data.extended_struct_variable),
            file=self.outFile
        )
        write(
            '        encoder->EncodeStructPtrPreamble(nullptr);',
            file=self.outFile
        )
        write('    }', file=self.outFile)
        write('}', file=self.outFile)



#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2019 Valve Corporation
# Copyright (c) 2018-2019 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from khronos_base_generator import write


class KhronosStructEncodersBodyGenerator():
    """KhronosStructEncodersBodyGenerator.
    Generates C++ functions for encoding Khronos API structures.
    """
    def skip_struct_type(self, struct_type):
        """Override as needed"""
        return False

    def write_encoder_content(self):
        api_data = self.get_api_data()
        for struct in self.get_all_filtered_struct_names():
            if self.skip_struct_type(struct):
                continue
            self.generate_struct_bodies(api_data, struct, self.all_struct_members[struct])

    def generate_struct_bodies(self, api_data, struct, struct_members):
        """Performs C++ code generation for the feature."""
        body = '\n'
        value_name = 'value'
        value_ref = value_name + '.'
        array_loop_specialization = None
        struct_type_var = self.get_struct_type_var_name()
        body += 'void EncodeStruct(ParameterEncoder* encoder, const {}& {})\n'.format(
            struct, value_name
        )
        body += '{\n'
        if struct in self.children_structs:
            body += self.make_child_struct_cast_switch(
                struct, value_name, struct_type_var
            )
            array_loop_specialization = self.make_child_loop_cast_switch(
                struct, struct_type_var
            )
        else:
            body += self.makeStructBody(api_data, struct, struct_members, value_ref)
        body += '}'
        if (array_loop_specialization):
            body += '\n\n' + array_loop_specialization
        write(body, file=self.outFile)

    def make_child_struct_cast_switch(self, parent_struct, value, struct_type_var):
        default_case = 'GFXRECON_LOG_WARNING("EncodeStruct({}): unrecognized child structure type %d", {}.{});'.format(
            parent_struct, value, struct_type_var
        )
        break_string = 'break;'
        switch_expression = 'value.{}'.format(struct_type_var)
        fn_emit_default = lambda parent_struct, value_name: [
            default_case, break_string
        ]
        fn_emit_case = lambda parent_struct, child_struct, child_enum, value_name: [
            f'const {child_struct}& child_value = reinterpret_cast<const {child_struct}&>({value_name});',
            f'EncodeStruct(encoder, child_value);', break_string
        ]
        return self.generate_child_struct_switch_statement(
            parent_struct, value, '    ', switch_expression, fn_emit_default,
            fn_emit_case
        )

    def make_child_loop_cast_switch(self, parent_struct, struct_type_var):
        func = 'EncodeStructArrayLoop'
        value = 'value'
        switch_expression = 'value->{}'.format(struct_type_var)
        default_case = 'GFXRECON_LOG_WARNING("{}: unrecognized child structure type %d", {}->{});'.format(
            func, value, struct_type_var
        )
        break_string = 'break;'
        fn_emit_default = lambda parent_struct, value_name: [
            default_case, break_string
        ]
        fn_emit_case = lambda parent_struct, child_struct, child_enum, value_name: [
            f'{func}<{child_struct}>(encoder, reinterpret_cast<const {child_struct} *>({value_name}), len);',
            break_string
        ]
        body = ''
        body += 'template <>\n'
        body += f'void EncodeStructArrayLoop<{parent_struct}>(ParameterEncoder* encoder, const {parent_struct}* {value}, size_t len)\n'
        body += '{\n'
        body += self.generate_child_struct_switch_statement(
            parent_struct, value, '    ', switch_expression, fn_emit_default,
            fn_emit_case
        )

        body += '}\n'
        return body

    def makeStructBody(self, api_data, name, values, prefix):
        """Command definition."""
        # Build array of lines for function body
        body = ''

        for value in values:
            # pNext fields require special treatment and are not processed by typename
            if self.is_extended_struct_definition(value):
                body += '    Encode{}Struct(encoder, {});\n'.format(
                    api_data.extended_struct_func_prefix, prefix + value.name
                )
            else:
                method_call = self.make_encoder_method_call(
                    name, value, values, prefix
                )
                body += '    {};\n'.format(method_call)

        return body



#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2019 Valve Corporation
# Copyright (c) 2018-2019 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from vulkan_base_generator import VulkanBaseGenerator, VulkanBaseGeneratorOptions, ValueInfo, write


class VulkanCommandBufferUtilBodyGeneratorOptions(VulkanBaseGeneratorOptions):
    """Options for generating a C++ class for Vulkan capture file replay."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        VulkanBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.extend((
            'generated/generated_vulkan_command_buffer_util.h',
            '',
            'encode/vulkan_handle_wrapper_util.h',
            'encode/vulkan_state_info.h',
        ))
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'encode'))
        self.begin_end_file_data.common_api_headers = []


class VulkanCommandBufferUtilBodyGenerator(VulkanBaseGenerator):
    """VulkanCommandBufferUtilBodyGenerator - subclass of VulkanBaseGenerator.
    Generates C++ member definitions for the VulkanReplayConsumer class responsible for
    replaying decoded Vulkan API call parameter data.
    Generate a C++ class for Vulkan capture file replay.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        VulkanBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        # The following functions require custom implementations
        self.customImplementationRequired = {
            'CmdPushDescriptorSetKHR'
        }


    def endFile(self):
        """Method override."""
        command_info = dict()  # Map of Vulkan commands to parameter info
        for cmd in self.get_all_filtered_cmd_names():
            command_info[cmd] = self.all_cmd_params[cmd]

        for cmd, info in command_info.items():
            wrapper_prefix = self.get_wrapper_prefix_from_command(cmd)
            if not cmd[2:] in self.customImplementationRequired:
                params = info[2]
                if params and params[0].base_type == 'VkCommandBuffer':
                    # Check for parameters with handle types, ignoring the first VkCommandBuffer parameter.
                    handles = self.get_param_list_handles(params[1:])

                    if (handles):
                        # Generate a function to build a list of handle types and values.
                        cmddef = '\n'
                        cmddef += 'void Track{}Handles({}::CommandBufferWrapper* wrapper, {})\n'.format(
                            cmd[2:], wrapper_prefix, self.get_arg_list(handles)
                        )
                        cmddef += '{\n'
                        indent = self.INDENT_SIZE * ' '
                        cmddef += indent + 'assert(wrapper != nullptr);\n'
                        cmddef += '\n'
                        for index, handle in enumerate(handles):
                            cmddef += self.insert_command_handle(
                                index, handle, indent=indent
                            )
                        cmddef += '}'

                        write(cmddef, file=self.outFile)

        self.newline()

        # Finish processing in superclass
        VulkanBaseGenerator.endFile(self)

    def need_feature_generation(self):
        """Indicates that the current feature has C++ code to generate."""
        if self.feature_cmd_params:
            return True
        return False

    def get_arg_list(self, values):
        args = []
        for value in values:
            if value.array_length:
                args.append('uint32_t {}'.format(value.array_length))
            args.append('{} {}'.format(value.full_type, value.name))
        return ', '.join(self.make_unique_list(args))

    def insert_command_handle(self, index, value, value_prefix='', indent=''):
        body = ''
        tail = ''
        index_name = None
        if (
            value.is_pointer or value.is_array
        ) and value.name != 'pnext_value':
            if index > 0:
                body += '\n'
            body += indent + 'if ({}{} != nullptr)\n'.format(
                value_prefix, value.name
            )
            body += indent + '{\n'
            tail = indent + '}\n' + tail
            indent += ' ' * self.INDENT_SIZE

            if value.is_array:
                index_name = '{}_index'.format(value.name)
                body += indent + 'for (uint32_t {i} = 0; {i} < {}{}; ++{i})\n'.format(
                    value_prefix, value.array_length, i=index_name
                )
                body += indent + '{\n'
                tail = indent + '}\n' + tail
                indent += ' ' * self.INDENT_SIZE

        if self.is_handle(value.base_type):
            type_enum_value = '{}Handle'.format(value.base_type[2:])
            wrapper_prefix = self.get_wrapper_prefix_from_type(value.base_type)
            value_name = value_prefix + value.name
            if value.is_array:
                value_name = '{}[{}]'.format(value_name, index_name)
            elif value.is_pointer:
                value_name = '(*{})'.format(value_name)
            body += indent + 'if({} != VK_NULL_HANDLE) wrapper->command_handles[vulkan_state_info::CommandHandleType::{}].insert(vulkan_wrappers::GetWrappedId<{}>({}));\n'.format(
                value_name, type_enum_value, wrapper_prefix + '::' + value.base_type[2:] + 'Wrapper' ,value_name
            )

        elif self.is_struct(
            value.base_type
        ) and (value.base_type in self.structs_with_handles):
            if value.is_array:
                access_operator = '[{}].'.format(index_name)
            elif value.is_pointer:
                access_operator = '->'
            else:
                access_operator = '.'

            for index, entry in enumerate(
                self.structs_with_handles[value.base_type]
            ):
                if entry.name == 'pNext':
                    ext_structs_with_handles = [
                        ext_struct for ext_struct in
                        self.registry.validextensionstructs[value.base_type]
                        if ext_struct in self.structs_with_handles
                    ]
                    if ext_structs_with_handles:
                        body += indent + 'auto pnext_header = reinterpret_cast<const VkBaseInStructure*>({}{}->pNext);\n'.format(
                            value_prefix, value.name
                        )
                        body += indent + 'while (pnext_header)\n'
                        body += indent + '{\n'
                        indent += ' ' * self.INDENT_SIZE
                        body += indent + 'switch (pnext_header->sType)\n'
                        body += indent + '{\n'
                        indent += ' ' * self.INDENT_SIZE
                        body += indent + 'default:\n'
                        indent += ' ' * self.INDENT_SIZE
                        body += indent + 'break;\n'
                        indent = indent[:-self.INDENT_SIZE]
                        for ext_struct in ext_structs_with_handles:
                            body += indent + 'case {}:\n'.format(
                                self.struct_type_names[ext_struct]
                            )
                            body += indent + '{\n'
                            indent += ' ' * self.INDENT_SIZE
                            body += indent + 'auto pnext_value = reinterpret_cast<const {}*>(pnext_header);\n'.format(
                                ext_struct
                            )
                            body += self.insert_command_handle(
                                index,
                                ValueInfo(
                                    'pnext_value', ext_struct,
                                    'const {} *'.format(ext_struct), 1
                                ),
                                '',
                                indent=indent
                            )
                            body += indent + 'break;\n'
                            indent = indent[:-self.INDENT_SIZE]
                            body += indent + '}\n'
                        indent = indent[:-self.INDENT_SIZE]
                        body += indent + '}\n'
                        body += indent + 'pnext_header = pnext_header->pNext;\n'
                        indent = indent[:-self.INDENT_SIZE]
                        body += indent + '}\n'
                else:
                    body += self.insert_command_handle(
                        index, entry,
                        value_prefix + value.name + access_operator, indent
                    )

        return body + tail



#!/usr/bin/python3
#
# Copyright 2018-2024, The Khronos Group Inc.
# Copyright (c) 2018-2019 Collabora, Ltd.
#
# SPDX-License-Identifier: Apache-2.0
#
# Author(s):    Rylie Pavlik <rylie.pavlik@collabora.com>
"Utilities for processing files."

from pathlib import Path


class LinewiseFileProcessor:
    """A base class for code that processes an input file (or file handle) one line at a time."""

    def __init__(self):
        self._lines = []
        self._line_num = 0
        self._next_line = None
        self._line = ''
        self._filename = Path()

    @property
    def filename(self):
        """The Path object of the currently processed file"""
        return self._filename

    @property
    def relative_filename(self):
        """The current file's Path relative to the current working directory"""
        return self.filename.relative_to(Path('.').resolve())

    @property
    def line(self):
        """The current line, including any trailing whitespace and the line ending."""
        return self._line

    @property
    def line_number(self):
        """Get 1-indexed line number."""
        return self._line_num

    @property
    def line_rstripped(self):
        """The current line without any trailing whitespace."""
        if self.line is None:
            return None
        return self.line.rstrip()

    @property
    def trailing_whitespace(self):
        """The trailing whitespace of the current line that gets removed when accessing rstrippedLine"""
        assert self.line_rstripped is not None
        non_whitespace_length = len(self.line_rstripped)
        return self.line[non_whitespace_length:]

    @property
    def next_line(self):
        """Peek at the next line, if any."""
        return self._next_line

    @property
    def next_line_rstripped(self):
        """Peek at the next line, if any, without any trailing whitespace."""
        if self.next_line is None:
            return None
        return self.next_line.rstrip()

    def get_preceding_line(self, relative_index=-1):
        """Retrieve the line at an line number at the given relative index, if one exists. Returns None if there is no line there."""
        if relative_index >= 0:
            raise RuntimeError(
                'relativeIndex must be negative, to retrieve a preceding line.')
        if relative_index + self.line_number <= 0:
            # There is no line at this index
            return None
        return self._lines[self.line_number + relative_index - 1]

    def get_preceding_lines(self, num):
        """Get *up to* the preceding num lines. Fewer may be returned if the requested number aren't available."""
        return self._lines[- (num + 1):-1]

    def process_line(self, line_num, line):
        """Implement in your subclass to handle each new line."""
        raise NotImplementedError

    def _process_file_handle(self, file_handle):
        # These are so we can process one line earlier than we're actually iterating thru.
        processing_line_num = None
        processing_line = None

        def do_process_line():
            self._line_num = processing_line_num
            self._line = processing_line
            if processing_line is not None:
                self._lines.append(processing_line)
                self.process_line(processing_line_num, processing_line)

        for line_num, line in enumerate(file_handle, 1):
            self._next_line = line
            do_process_line()
            processing_line_num = line_num
            processing_line = line

        # Finally process the left-over line
        self._next_line = None
        do_process_line()

    def process_file(self, filename, file_handle=None):
        """Main entry point - call with a filename and optionally the file handle to read from."""
        if isinstance(filename, str):
            filename = Path(filename).resolve()

        self._filename = filename

        if file_handle:
            self._process_file_handle(file_handle)
        else:
            with self._filename.open('r', encoding='utf-8') as f:
                self._process_file_handle(f)



"""Utility functions not closely tied to other spec_tools types."""
# Copyright (c) 2018-2019 Collabora, Ltd.
# Copyright 2013-2025 The Khronos Group Inc.
#
# SPDX-License-Identifier: Apache-2.0


def getElemName(elem, default=None):
    """Get the name associated with an element, either a name child or name attribute."""
    name_elem = elem.find('name')
    if name_elem is not None:
        return name_elem.text
    # Fallback if there is no child.
    return elem.get('name', default)


def getElemType(elem, default=None):
    """Get the type associated with an element, either a type child or type attribute."""
    type_elem = elem.find('type')
    if type_elem is not None:
        return type_elem.text
    # Fallback if there is no child.
    return elem.get('type', default)


def findFirstWithPredicate(collection, pred):
    """Return the first element that satisfies the predicate, or None if none exist.

    NOTE: Some places where this is used might be better served by changing to a dictionary.
    """
    for elt in collection:
        if pred(elt):
            return elt
    return None


def findNamedElem(elems, name):
    """Traverse a collection of elements with 'name' nodes or attributes, looking for and returning one with the right name.

    NOTE: Many places where this is used might be better served by changing to a dictionary.
    """
    return findFirstWithPredicate(elems, lambda elem: getElemName(elem) == name)


def findTypedElem(elems, typename):
    """Traverse a collection of elements with 'type' nodes or attributes, looking for and returning one with the right typename.

    NOTE: Many places where this is used might be better served by changing to a dictionary.
    """
    return findFirstWithPredicate(elems, lambda elem: getElemType(elem) == typename)


def findNamedObject(collection, name):
    """Traverse a collection of elements with 'name' attributes, looking for and returning one with the right name.

    NOTE: Many places where this is used might be better served by changing to a dictionary.
    """
    return findFirstWithPredicate(collection, lambda elt: elt.name == name)



#!/usr/bin/python3 -i
#
# Copyright (c) 2022-2023 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from openxr_base_generator import *
from khronos_enum_to_json_body_generator import KhronosEnumToJsonBodyGenerator
from reformat_code import format_cpp_code


class OpenXrEnumToJsonBodyGeneratorOptions(OpenXrBaseGeneratorOptions):
    """Options for generating C++ functions for OpenXr FieldToJson() functions"""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefixText='',
        protectFile=False,
        protectFeature=True,
        extra_headers=[]
    ):
        OpenXrBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefixText,
            protectFile,
            protectFeature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.extend((
            'generated_openxr_enum_to_json.h',
            'util/to_string.h',
        ))
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'decode'))
        self.begin_end_file_data.common_api_headers = []


# OpenXrEnumToStringBodyGenerator - subclass of OpenXrBaseGenerator.
# Generates C++ functions for stringifying OpenXr API enums.
class OpenXrEnumToJsonBodyGenerator(OpenXrBaseGenerator, KhronosEnumToJsonBodyGenerator):
    """Generate C++ functions for OpenXr FieldToJson() functions"""

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        OpenXrBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    def skip_generating_enum_to_json_for_type(self, type):
        return type.startswith('Vk') or 'Bits' in type

    # Method override
    def endFile(self):
        KhronosEnumToJsonBodyGenerator.make_decls(self)

        # Finish processing in superclass
        OpenXrBaseGenerator.endFile(self)
